/* v0.1 w. ebisuzaki
 *
 * routines to make bds and bms
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>
#include <limits.h>
#include "gribw.h"

unsigned char *mk_void_BMS(unsigned char *pds) {
    clr_HasBMS(pds);
    return NULL;
}


unsigned char *mk_BMS(unsigned char *pds, float *bindata, int *n, float undef_low, float undef_hi) {
    int bms_size, i, nn, start;
    unsigned int imask, c;
    unsigned char *bms, *cbits;

    if (undef_low > undef_hi) {
        clr_HasBMS(pds);
        return NULL;
    }

    nn = *n;
    bms_size = 6 + (nn + 7) / 8;

    if (bms_size & 1) bms_size++;
    if ((bms = (unsigned char *) malloc(bms_size * sizeof(char))) == NULL) {
	fprintf(stderr,"*** not enough memory: mk_bms ***\n");
	exit(8);
    }

    set_int3(bms, bms_size);
    bms[3] = bms_size*8 - 6*8 - nn;
    bms[4] = bms[5] = bms[bms_size-1] = 0;

    cbits = bms + 6;
    c = start = 0;
    imask = 128;
    for (i = 0; i < nn; i++) {
	if (bindata[i] < undef_low || bindata[i] > undef_hi) {
	    c += imask;
	    bindata[start++] = bindata[i];
	}
	if ((imask >>= 1) == 0) {
	    *cbits++ = c;
	    c = 0;
	    imask = 128;
	}
    }
    if (imask != 128) *cbits = c;
    if (nn == start) {
	free(bms);
        clr_HasBMS(pds);
        return NULL;
    }
    set_HasBMS(pds);
    *n = start;
    return bms;
}

/* assumed that all the undefined values have been eliminated
 * from bindata
 *
 * get scaling from pds .. no binary scaling
 *
 * note bindata is destroyed
 *
 * v0.1
 */

unsigned char *mk_BDS(unsigned char *pds, float *bindata, int n) {

    float min_value, max_value;
    int imin, imax;

    double scale;
    int i, nbits, bds_size, unused_bits, *idata;
    unsigned char *bds;

    /* scale the data */
    scale = int_power(10.0, get_DecScale(pds));
    for (i = 0; i < n; i++) {
	bindata[i] *= scale;
    }

    /* find min value */
    for (max_value = min_value = bindata[0], i = 1; i < n; i++) {
	if (min_value > bindata[i]) min_value = bindata[i];
	if (max_value < bindata[i]) max_value = bindata[i];
    }

    if (max_value > INT_MAX | min_value < INT_MIN) {
	fprintf(stderr,"*** the data cannot be stored as an integer ***\n");
	fprintf(stderr,"*** you must have bad data!!! ***\n");
	fprintf(stderr,"*** check units or variable type ***\n");
	fprintf(stderr,"min value: %g max value %g\n",min_value/scale,
		(min_value+max_value)/scale);
	fprintf(stderr,"scaling factor: %g %d\n", scale, get_DecScale(pds));

	exit(8);
    }

    imin = min_value = floor(min_value+0.5);
    imax = floor(max_value+0.5) - imin;

    nbits = 0;
    while (imax) {
	nbits++;
	imax >>= 1;
    }
    bds_size = 11 + (nbits*n + 7)/8;
    if (bds_size % 2) bds_size++;
    unused_bits = bds_size*8 - 11*8 - nbits*n;
    if (unused_bits > 15 || unused_bits < 0) {
        fprintf(stderr,"programming error! unused bits\n");
        exit(8);
    }
 
    if ((bds = (unsigned char *) malloc(bds_size)) == NULL) { 
	fprintf(stderr,"*** not enough memory ***\n");
	exit(8);
    }

    /* want to allocate idata after bds .. for cray */

    if ((idata = (int *) malloc(n * sizeof(int))) == NULL) {
	fprintf(stderr,"*** not enough memory ***\n");
	exit(8);
    }

    for (i = 0; i < n; i++) {
	idata[i] = floor(bindata[i]+0.5) - imin;
    }

    /* now to fill in the bds */

    set_int3(bds, bds_size);
    bds[3] = 0;	/* flags */
    bds[4] = 0;	/* binary scale factor*/
    bds[5] = 0;
    flt2ibm(min_value, bds+6);	/* reference value */
    bds[10] = nbits;
    bds[3] = bds[3] | (unused_bits & 15);
    bds[bds_size-1] = 0;

    /* fill in the data */
    list2bitstream(idata, bds+11, n, nbits);
    free(idata);
    return bds;
}


